﻿using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.MetaOptimization {
  /// <summary>
  /// TODO 
  /// </summary>
  [Item("BestParameterConfigurationAnalyzer", "")]
  [StorableClass]
  public sealed class BestParameterConfigurationAnalyzer : SingleSuccessorOperator, IAnalyzer {

    public ScopeTreeLookupParameter<ParameterConfigurationTree> ParameterConfigurationParameter {
      get { return (ScopeTreeLookupParameter<ParameterConfigurationTree>)Parameters["ParameterConfigurationTree"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public LookupParameter<IRun> BestSolutionParameter {
      get { return (LookupParameter<IRun>)Parameters["BestSolution"]; }
    }
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public LookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    public LookupParameter<IRun> BestKnownSolutionParameter {
      get { return (LookupParameter<IRun>)Parameters["BestKnownSolution"]; }
    }
    public LookupParameter<RunCollection> PopulationParameter {
      get { return (LookupParameter<RunCollection>)Parameters["Population"]; }
    }
    public LookupParameter<ConstrainedItemList<IProblem>> ProblemsParameter {
      get { return (LookupParameter<ConstrainedItemList<IProblem>>)Parameters[MetaOptimizationProblem.ProblemsParameterName]; }
    }
    public LookupParameter<BoolValue> MaximizationParameter {
      get { return (LookupParameter<BoolValue>)Parameters["Maximization"]; }
    }

    public BestParameterConfigurationAnalyzer()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<ParameterConfigurationTree>("ParameterConfigurationTree", ""));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", ""));
      Parameters.Add(new LookupParameter<IRun>("BestSolution", ""));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", ""));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", ""));
      Parameters.Add(new LookupParameter<IRun>("BestKnownSolution", ""));
      Parameters.Add(new LookupParameter<RunCollection>("Population", ""));
      Parameters.Add(new LookupParameter<ConstrainedItemList<IProblem>>(MetaOptimizationProblem.ProblemsParameterName));
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", "Set to false if the problem should be minimized."));
    }

    [StorableConstructor]
    private BestParameterConfigurationAnalyzer(bool deserializing) : base(deserializing) { }
    private BestParameterConfigurationAnalyzer(BestParameterConfigurationAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestParameterConfigurationAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      ItemArray<DoubleValue> qualities = QualityParameter.ActualValue;
      ResultCollection results = ResultsParameter.ActualValue;
      DoubleValue bestKnownQuality = BestKnownQualityParameter.ActualValue;
      ItemArray<ParameterConfigurationTree> parameterTrees = ParameterConfigurationParameter.ActualValue;
      bool maximization = MaximizationParameter.ActualValue.Value;

      int idxBest;
      if (maximization)
        idxBest = qualities.Select((x, index) => new { index, x.Value }).OrderBy(x => x.Value).Last().index;
      else
        idxBest = qualities.Select((x, index) => new { index, x.Value }).OrderBy(x => x.Value).First().index;

      ParameterConfigurationTree best = (ParameterConfigurationTree)parameterTrees[idxBest];
      IRun bestRun = new Run();
      best.CollectResultValues(bestRun.Results);
      best.CollectParameterValues(bestRun.Parameters);

      if (bestKnownQuality == null || 
        (!maximization && (qualities[idxBest].Value < bestKnownQuality.Value) ||
        (maximization && (qualities[idxBest].Value > bestKnownQuality.Value)))) {
        BestKnownQualityParameter.ActualValue = new DoubleValue(qualities[idxBest].Value);

        BestKnownSolutionParameter.ActualValue = bestRun;
      }

      if (BestSolutionParameter.ActualValue == null) {
        BestSolutionParameter.ActualValue = bestRun;
        results.Add(new Result("Best Parameter Settings", bestRun));
      } else {
        BestSolutionParameter.ActualValue = bestRun;
        results["Best Parameter Settings"].Value = bestRun;
      }

      // population
      int i = 0;
      RunCollection rc = new RunCollection();
      foreach (ParameterConfigurationTree pt in (maximization ? parameterTrees.OrderByDescending(x => x.Quality) : parameterTrees.OrderBy(x => x.Quality))) {
        rc.Add(pt.ToRun(string.Format("Individual {0} ({1})", i, pt.ParameterInfoString)));
        i++;
      }
      if (PopulationParameter.ActualValue == null) {
        PopulationParameter.ActualValue = rc;
        results.Add(new Result("Population", rc));
      } else {
        PopulationParameter.ActualValue = rc;
        results["Population"].Value = rc;
      }

      return base.Apply();
    }
  }
}
